package geniusweb.partiesserver.repository;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Date;
import java.util.UUID;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonAutoDetect.Visibility;

import geniusweb.actions.PartyId;
import geniusweb.inform.Inform;
import geniusweb.partiesserver.RunningPartiesUpdater;
import geniusweb.partiesserver.websocket.PartySocket;
import geniusweb.party.Party;
import tudelft.utilities.repository.RepoElement;

/**
 * Contains a running party and additional info about it like time constraints
 * and connections. Adds a start/end times and generates party ID. Also stores
 * original filename of the party.
 * 
 * This object can not be deserialized here because running parties have to be
 * created.
 */
@JsonAutoDetect(fieldVisibility = Visibility.ANY, getterVisibility = Visibility.NONE, setterVisibility = Visibility.NONE)
public class RunningParty implements RepoElement<PartyId> {
	private static int nr = 1; // counter for generating unique new parties.
	private static String uniqueid = getUniqueMachineID(); // world-wide unique
															// ID for this
															// number. For ID
															// generation.

	private final PartyId id;
	private final Date startDate;
	private final Date endDate; // agent will be declared dead and removed after
								// this
	private final String name;
	/*
	 * set when someone contacts our socket. null if not set.
	 */
	private final transient PartySocket connection;

	/**
	 * The actually running party
	 */
	private transient final Party party;

	/**
	 * Only available for testing.
	 * 
	 * @param party the new party
	 * @param id    the new id for the party.
	 * @param name  the name, this should match the filename
	 * @param start the start date for this party. Usually set to current time.
	 * @param end   the end date for this party
	 * @param conn  the connection to use, or null if no connection is available
	 *              yet.
	 */
	protected RunningParty(Party party, PartyId id, String name, Date start,
			Date end, PartySocket conn) {
		if (party == null || id == null || name == null || start == null
				|| end == null) {
			throw new NullPointerException("arguments must not be null");
		}
		this.party = party;
		this.id = id;
		this.name = name;
		this.startDate = start;
		this.endDate = end;
		this.connection = conn;
	}

	/**
	 * @param availableparty the {@link AvailableParty}
	 * @param maxlifetimems  the maximum life time for this party (ms). The
	 *                       party will be killed and removed after this point.
	 *                       This value should be set to the maximum negotiation
	 *                       time plus the worst time to actual start up of the
	 *                       negotiation
	 * @return a new RunningParty being an instance of availableparty
	 * @throws InstantiationFailedException if class can not be instantiated
	 * 
	 */
	public static RunningParty create(AvailableParty availableparty,
			long maxlifetimems) throws InstantiationFailedException {
		if (availableparty == null) {
			throw new IllegalArgumentException(
					"availableparty must not be null");
		}
		return create(availableparty.getInstance(), availableparty.getName(),
				maxlifetimems);
	}

	/**
	 * Create a running party.
	 * 
	 * @param party        the {@link Party} that is loaded and ready to run.
	 * @param name         The name, this should match the filename in the repo
	 *                     (without the .jar).
	 * @param maxRunTimeMS the maximum runtime for this party (ms). The
	 *                     {@link RunningPartiesUpdater} will keep an eye on the
	 *                     time and handle removal after time is up.
	 * @return new RunningParty created from the Party instance.
	 * 
	 */
	private static RunningParty create(Party party, String name,
			long maxRunTimeMS) {
		if (maxRunTimeMS <= 0) {
			throw new IllegalArgumentException("runtime must be >0");
		}
		Date now = new Date();
		return new RunningParty(party,
				new PartyId(squash1(name) + "_" + uniqueid + "_" + (nr++)),
				name, now, new Date(now.getTime() + maxRunTimeMS), null);
	}

	/**
	 * @return unique ID for this running party
	 */
	@Override
	public PartyId getID() {
		return id;
	}

	/**
	 * @return the implementation of this party
	 */
	public Party getParty() {
		return party;
	}

	/**
	 * @return The name, this should match the filename in the repo (without the
	 *         .jar).
	 * 
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return the start {@link Date} of this runningparty.
	 */
	public Date getStartDate() {
		return startDate;
	}

	/**
	 * 
	 * @return the end date/ deadline for this party
	 */
	public Date getEndDate() {
		return endDate;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((id == null) ? 0 : id.hashCode());
		result = prime * result + ((name == null) ? 0 : name.hashCode());
		result = prime * result + ((party == null) ? 0 : party.hashCode());
		result = prime * result
				+ ((startDate == null) ? 0 : startDate.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		RunningParty other = (RunningParty) obj;
		if (id == null) {
			if (other.id != null)
				return false;
		} else if (!id.equals(other.id))
			return false;
		if (name == null) {
			if (other.name != null)
				return false;
		} else if (!name.equals(other.name))
			return false;
		if (party == null) {
			if (other.party != null)
				return false;
		} else if (!party.equals(other.party))
			return false;
		if (startDate == null) {
			if (other.startDate != null)
				return false;
		} else if (!startDate.equals(other.startDate))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "RunningParty[" + id + "," + startDate + party.getClass() + ","
				+ name + "]";
	}

	/**
	 * Pass {@link Inform} into the running party. You must have called
	 * {@link #withConnection(PartySocket)} before calling this.
	 * 
	 * @param info the {@link Inform} to be sent to the {@link #connection}
	 */
	public void inform(Inform info) {
		connection.notifyListeners(info);
	}

	/**
	 * Attach RunningParty to a connection. We allow at most 1 connection.
	 * 
	 * @param newconn the new connection that has been made with this party.
	 * @return new RunningParty. This must be inserted into the
	 *         {@link RunningPartiesRepo} by the caller.
	 * @throws IllegalStateException if the party is already connected.
	 */
	public RunningParty withConnection(PartySocket newconn) {
		if (newconn == null)
			throw new NullPointerException("newconn must be not null");
		if (this.connection != null) {
			throw new IllegalStateException(
					"RunningParty is already connected");
		}
		party.connect(newconn);
		return new RunningParty(party, id, name, startDate, endDate, newconn);
	}

	/**
	 * 
	 * @param terminationTime the new termination time
	 * @return new RunningParty with changed end time (all other things
	 *         unchanged)
	 */
	public RunningParty withEndTime(Date terminationTime) {
		return new RunningParty(party, id, name, startDate, terminationTime,
				connection);
	}

	/**
	 * Get a id that is world-wide unique. We use the IP number here instead of
	 * UUID because UUID usually is just a random number which is (1) long (22
	 * bytes) (2) not guaranteed unique.
	 */
	private static String getUniqueMachineID() {
		if (uniqueid == null) {
			try {
				uniqueid = InetAddress.getLocalHost().getHostAddress();
			} catch (UnknownHostException e1) {
				uniqueid = UUID.randomUUID().toString(); // fallback, should
															// never be needed
			}
			uniqueid = squash(uniqueid);
		}
		return uniqueid;
	}

	/**
	 * 
	 * @param string
	 * @return string with all non-word characters replaced with "_" so that the
	 *         result is a "word" ( [a-zA-Z_0-9] ).
	 */
	protected static String squash(String string) {
		return string.replaceAll("\\W", "_");
	}

	/**
	 * @param string
	 * @return as squash, but additionally if the name does not start with a
	 *         letter we prefix the string with 'p'
	 */
	protected static String squash1(String string) {
		String res = squash(string);
		if (res.matches("[a-zA-Z]\\w*"))
			return res;
		return "p" + res;
	}

}
